@@ -1,9 +1,8 @@
 //! Shuttle's implementation of [`std::sync`].
 
+pub mod atomic;
 mod barrier;
 mod condvar;
-
-/// Multi-producer, single-consumer FIFO queue communication primitives.
 pub mod mpsc;
 mod mutex;
 mod rwlock;
