@@ -1,4 +1,6 @@
+mod atomic;
 mod barrier;
+mod clocks;
 mod condvar;
 mod config;
 mod dfs;
@@ -6,6 +8,7 @@ mod execution;
 mod metrics;
 mod mpsc;
 mod mutex;
+mod panic;
 mod pct;
 mod portfolio;
 mod replay;
